@@ -1,20 +1,5 @@
 # -*- coding: utf-8 -*-
-import os
-import json
-import pkg_resources
-
-import numpy
-
-from .units import conversion_facs_frequency
-from .units import conversion_facs_energy
-
-from .singleton import Singleton
-
-
-class Manager(metaclass=Singleton):
-    """ Main package Manager
-
-
+"""
     This class handles several important package wide tasks:
 
     1) Usage of units across objects storing data
@@ -44,29 +29,78 @@ class Manager(metaclass=Singleton):
         
     units_repre_latex : dictionary
         dictionary of latex prepresentations of available units
+
+
+
+    Units Management
+    ----------------
+    Units management is performed for all classes derived from
+    quantarhei.managers.UnitsManaged class.
+
+
+    Basis Conversion Management
+    ---------------------------
+    Units management is performed for all classes derived from
+    quantarhei.managers.BasisManaged class.
+
+    Basis management works like this: when an class is defined, and its
+    property needs to be basis managed, one should use a predefined type
+    `basis_managed_array_property`
+
+
+
     
+
+"""
+import os
+import warnings
+
+#
+# This stops future warnings, notably those in h5py library
+# FIXME: remove this in "future"
+#
+warnings.simplefilter(action='ignore', category=FutureWarning)
+
+import json
+import pkg_resources
+
+import numpy
+
+from .units import conversion_facs_frequency
+from .units import conversion_facs_energy
+from .units import conversion_facs_length
+
+from .singleton import Singleton
+
+from .numconf import NumConf
+from .logconf import LogConf
+from .genconf import GenConf
+
+class Manager(metaclass=Singleton):
+    """ Main package Manager
+
     """
-    
 
-
-    #version = "0.0.11"
-    version = pkg_resources.require("quantarhei")[0].version
+    version = "0.0.57"
 
     # hard wired unit options
     allowed_utypes = ["energy",
                       "frequency",
                       "dipolemoment",
-                      "temperature",
-                      "time"]
+                      "temperature",   
+                      "time",
+                      "length"]
 
-    units = {"energy"       : ["2pi/fs", "int", "1/cm", "eV", "meV", "THz",
-                               "J", "SI", "nm"],
-             "frequency"    : ["2pi/fs", "int", "1/cm", "THz","Hz","SI"],
-             "dipolemoment" : ["Debye"],
-             "temperature"  : ["2pi/fs", "int", "Kelvin", "Celsius",
+    units = {"energy"       : ["1/fs", "int", "1/cm", "eV", "meV", "THz",
+                               "J", "SI", "nm", "Ha", "a.u."],
+             "frequency"    : ["1/fs", "int", "1/cm", "THz", "Hz", "SI",
+                               "nm", "Ha", "a.u."],
+             "dipolemoment" : ["Debye", "a.u"],
+             "temperature"  : ["1/fs", "int", "Kelvin", "Celsius",
                                "1/cm", "eV", "meV", "Thz", "SI"],
              "time"         : ["fs", "int", "as", "ps", "ns", "Ms","ms",
-                               "s", "SI"]}
+                               "s", "SI"],
+             "length"       : ["int", "A", "nm", "Bohr", "a.u.", "m", "SI"]}
 
     units_repre = {"Kelvin":"K",
                    "Celsius":"C",
@@ -74,10 +108,12 @@ class Manager(metaclass=Singleton):
                    "1/cm":"1/cm",
                    "THz":"THz",
                    "eV":"eV",
-                   "2pi/fs":"2pi/fs",
+                   "1/fs":"1/fs",
                    "int":"2pi/fs",
                    "meV":"meV",
-                   "nm":"nm"}
+                   "nm":"nm",    
+                   "Ha":"Ha",
+                   "a.u.":"a.u."}   
                    
     units_repre_latex = {"Kelvin":"K",
                    "Celsius":"C",
@@ -85,12 +121,21 @@ class Manager(metaclass=Singleton):
                    "1/cm":"cm$^-1$",
                    "THz":"THz",
                    "eV":"eV",
-                   "2pi/fs":"rad$\cdot$fs$^{-1}$",
+                   "1/fs":"fs$^{-1}$",
                    "meV":"meV",
-                   "nm":"nm"}                  
+                   "nm":"nm",
+                   "Ha":"Ha",
+                   "a.u.":"a.u."}                  
 
     def __init__(self):
-
+        
+        try:
+            # this is numpy 1.14
+            numpy.set_printoptions(precision=8, sign=' ', legacy='1.13')
+        except:
+            # before there was no `sign` parameters
+            numpy.set_printoptions(precision=8)
+        
         self.current_units = {}
 
         # main configuration file
@@ -123,6 +168,9 @@ class Manager(metaclass=Singleton):
             # load the main configuration file
             with open(self.cfile, 'r') as f:
                 self.main_conf = json.load(f)
+                
+                
+        self.current_basis_operator = None
 
 
 
@@ -132,16 +180,16 @@ class Manager(metaclass=Singleton):
         #
 
         # internal units are hardwired
-        self.internal_units = {"energy":"2pi/fs", "frequency":"2pi/fs",
+        self.internal_units = {"energy":"1/fs", "frequency":"1/fs",
                                "dipolemoment":"Debye",
-                               "temperature":"Kelvin"}
+                               "temperature":"Kelvin", "length":"A"}
 
         # current units are read from conf file
         if not exists:
             # set hard wired defaults and save them
-            self.current_units = {"energy":"2pi/fs", "frequency":"2pi/fs",
+            self.current_units = {"energy":"1/fs", "frequency":"1/fs",
                                   "dipolemoment":"Debye",
-                                  "temperature":"Kelvin"}
+                                  "temperature":"Kelvin", "length":"A"}
 
 
             # save them
@@ -160,26 +208,41 @@ class Manager(metaclass=Singleton):
             "secular-standard-Redfield-rates":"redfield.ssRedfieldRateMatrix"
 
            }
+        
+        #
+        #  All available implementations
+        #
         self.all_implementations = {
-            
-           "redfield.ssRedfieldRateMatrix": {
+           "redfieldrates.ssRedfieldRateMatrix": {
                '0':"quantarhei.implementations.python",
                '1':"quantarhei.implementations.cython"
                }
-           
            }
+            
+        self.all_implementations["redfieldtensor.ssRedfieldTensor"] = \
+            {'0':"quantarhei.implementations.python",
+             '1':"quantarhei.implementations.cython"}
+            
                 
         self.default_implementations = {
-            "redfield.ssRedfieldRateMatrix":'0'
+            "redfieldrates.ssRedfieldRateMatrix":'0',
+            "redfieldtensor.ssRedfieldRateTensor":'0'
             }
 
         self.optimal_implementations = {
-            "redfield.ssRedfieldRateMatrix":'1'
+            "redfieldrates.ssRedfieldRateMatrix":'1'
             }
                 
         self.current_implementations = {
-            "redfield.ssRedfieldRateMatrix":'0'        
+            "redfieldrates.ssRedfieldRateMatrix":'0',
+            "redfieldtensor.ssRedfieldRateTensor":'0'        
             }
+        
+        self.parallel_implementations = {}
+        self.parallel_implementations["redfieldrates."
+                                 +"ssRedfieldRateMatrix"] = \
+            {'0':"quantarhei.implementations.python.parallel",
+             '1':"quantarhei.implementations.cython.parallel"}
                 
         if not exists:
             
@@ -190,7 +253,7 @@ class Manager(metaclass=Singleton):
             self.load_implementations()
             
             
-        self.change_implementation_at_runtime = False
+        self.change_implementation_at_runtime = True
         
         self.basis_stack = []
         self.basis_stack.append(0)
@@ -205,7 +268,132 @@ class Manager(metaclass=Singleton):
         
         self.save_dict = {}
         
+        
+        #
+        # Configuration controlable from qrhei (conf file and qrhei script)
+        #
+        self.num_conf = NumConf()
+        
+        
+        self.log_conf = LogConf()
+        
+        self.use_pytorch = False
+        self.use_gpu = False
+
+
+        self.gen_conf = GenConf()        
+        
+        #
+        # Read central configuration from ./quantarhei directory
+        #
+        
+         
+        
+        #
+        # Read local user config file (this will only be done on request)
+        #
+        # self._read_uconf()
+        
+        
+        
+        #
+        # Initialization of parallel environment
+        #
+        try:
+            from .parallel import DistributedConfiguration
+            #from .parallel import start_parallel_region
+            dc = DistributedConfiguration()
+            self.parallel_conf = dc
+            dc.start_parallel_region()
             
+            # this must be put into qrhei script !!!
+            #if dc.rank != 0:
+            #    self.log_conf.verbosity -= 2
+            #    self.log_conf.fverbosity -= 2
+            #print(dc.rank, self.log_conf.verbosity)
+        except:
+            self.parallel_conf = None
+            
+        
+    def __del__(self):
+        """Closes parallel environment if needed
+        
+        """
+        
+        if self.parallel_conf is not None:
+            #from .parallel import close_parallel_region
+            self.parallel_conf.finish_parallel_region()
+
+            
+        
+        
+    def load_conf(self):
+        """Loads configuration file
+        
+        This is to be called in scripts and notebooks
+        
+        """
+        self._read_uconf()
+        
+        
+    def _read_uconf(self):
+        """Reads used defined local config file
+        
+        From Stackoverflow recipe:
+            https://stackoverflow.com/questions/67631/how-to-import-a-module-given-the-full-path
+            
+        
+        """
+        fname = self.gen_conf.conf_file_name
+        fdir = self.gen_conf.conf_file_path
+        fpath = os.path.join(fdir, fname)
+        
+        from pathlib import Path
+        cfile = Path(fpath)   
+        
+        if cfile.exists() & cfile.is_file():
+
+            self._load_uconf(fpath)
+                
+        else:
+            if cfile.exists():
+                raise Exception("Configuration file "+fpath+" seems to exits"+
+                                " but it is not a file")
+            else:
+                print("Warning: Configuration file "+fpath+" does not exit")
+                print("Warning: Placing a default configuration are using it")
+                
+                import pkg_resources
+
+                resource_package = "quantarhei"  # Could be any module/package name
+                resource_path = '/'.join(('core', 'conf', 'qrhei.py')) 
+                content = pkg_resources.resource_string(resource_package,
+                                                        resource_path)
+
+                with open(fpath, "w") as f:
+                    f.write(content.decode("utf-8"))
+                    
+                self._load_uconf(fpath)
+        
+        #printlog("Configuration file: ", fpath, "loaded", loglevel=9)        
+       
+    def _load_uconf(self, fpath):
+        """
+        
+        """
+        #print("Conf path: ", os.path.abspath(fpath))
+        try:
+            import importlib.util
+            spec = importlib.util.spec_from_file_location("qrconf", fpath)
+            foo = importlib.util.module_from_spec(spec)
+            spec.loader.exec_module(foo)    
+            #print("Configuring Manager:")
+            #print(self)
+            foo.configure(self)
+            #print("..done")
+        except:
+            raise Exception()        
+        
     def save_settings(self):
 
         # main configuration file
@@ -257,7 +445,26 @@ class Manager(metaclass=Singleton):
         with open(units_file,'r') as f:
             self.current_units = json.load(f)       
     
+    def get_real_type(self):
+        """Returns default numpy float type
+        
+        """
+        import numpy
+        return numpy.float64
+    
+    def get_complex_type(self):
+        """Returns default numpy complex type
+        
+        """
+        import numpy
+        return numpy.complex128
 
+
+    def store_current_basis_operator(self, op):
+        self.current_basis_operator = op
+        
+    def remove_current_basis_operator(self):
+        self.current_basis_operator = None
         
     def unit_repr(self,utype="energy",mode="current"):
         """Returns a string representing the currently used units
@@ -297,11 +504,14 @@ class Manager(metaclass=Singleton):
             
             
             
-    def set_current_units(self,utype,units):
+    def set_current_units(self, utype, units):
         """Sets current units
         
         
         """
+        self._saved_units = {}
+        self._saved_units[utype] = self.get_current_units(utype)
+        
         if utype in self.allowed_utypes:
             if units in self.units[utype]:
                 self.current_units[utype] = units
@@ -309,8 +519,27 @@ class Manager(metaclass=Singleton):
                 raise Exception("Unknown units of %s" % utype)
         else:
             raise Exception("Unknown type of units")
+        
+    def unset_current_units(self, utype):
+        """Restores previously saved units of a given type
+        
+        """
+        try:
+            cunits = self._saved_units[utype]
+        except KeyError:
+            raise Exception("Units to restore not found")
             
-    def get_current_units(self,utype):
+        if utype in self.allowed_utypes:
+            if cunits in self.units[utype]:
+                self.current_units[utype] = cunits
+            else:
+                raise Exception("Unknown units of %s" % utype)
+        else:
+            raise Exception("Unknown type of units")
+        
+        
+        
+    def get_current_units(self, utype):
         """
         
         """
@@ -351,7 +580,7 @@ class Manager(metaclass=Singleton):
         """Convert energy from currently used units to internal units
         
         Parameters
-        ==========
+        ----------
 
         val : number, array, list, tuple of numbers
             values to convert            
@@ -377,10 +606,10 @@ class Manager(metaclass=Singleton):
         
             
     def convert_energy_2_current_u(self,val):
-        """Convert energy from internal units to currently used units
+        """Converts energy from internal units to currently used units
         
         Parameters
-        ==========
+        ----------
 
         val : number, array, list, tuple of numbers
             values to convert            
@@ -403,10 +632,10 @@ class Manager(metaclass=Singleton):
         
 
     def convert_frequency_2_internal_u(self,val):
-        """Convert frequency from currently used units to internal units
+        """Converts frequency from currently used units to internal units
         
         Parameters
-        ==========
+        ----------
 
         val : number, array, list, tuple of numbers
             values to convert            
@@ -416,10 +645,10 @@ class Manager(metaclass=Singleton):
 
         
     def convert_frequency_2_current_u(self,val):   
-        """Convert frequency from internal units to currently used units
+        """Converts frequency from internal units to currently used units
         
         Parameters
-        ==========
+        ----------
 
         val : number, array, list, tuple of numbers
             values to convert            
@@ -427,8 +656,32 @@ class Manager(metaclass=Singleton):
         """
         return val/conversion_facs_frequency[self.current_units["frequency"]] 
         
+
+    def convert_length_2_internal_u(self,val):
+        """Converts length from currently used units to internal units
         
+        Parameters
+        ----------
+
+        val : number, array, list, tuple of numbers
+            values to convert            
         
+        """
+        return val*conversion_facs_length[self.current_units["length"]]        
+
+
+    def convert_length_2_current_u(self,val):   
+        """Converts frequency from internal units to currently used units
+        
+        Parameters
+        ----------
+
+        val : number, array, list, tuple of numbers
+            values to convert            
+        
+        """
+        return val/conversion_facs_length[self.current_units["length"]]   
+      
             
     def get_implementation_prefix(self,package="",taskname=""):
         #default_imp_prefix = "quantarhei.implementations.python"
@@ -456,20 +709,16 @@ class Manager(metaclass=Singleton):
         whichone = self.current_implementations[imp_id]
         return self.all_implementations[imp_id][str(whichone)]
         
-    def set_current_implementation(self,imp,choice):
+    def set_current_implementation(self, imp, choice):
         imp_id = self.implementation_points[imp]
         self.current_implementations[imp_id] = choice
-        
+
     def register_implementation(self,imp_point,prefix,asint=None):
         pass
     
     def commit_implementation(self,imp_point,prefix,asint=None):
         pass
     
-    
-    
-    
-    
     def get_current_basis(self):
         """Returns the current basis id
         
@@ -547,26 +796,12 @@ class Manager(metaclass=Singleton):
         return self.parallel_conf
 
 
-"""
-
-    Units Management
-    ----------------
-    Units management is performed for all classes derived from
-    quantarhei.managers.UnitsManaged class.
-
-
-    Basis Conversion Management
-    ---------------------------
-    Units management is performed for all classes derived from
-    quantarhei.managers.BasisManaged class.
-
-    Basis management works like this: when an class is defined, and its
-    property needs to be basis managed, one should use a predefined type
-    `basis_managed_array_property`
 
 
 
-"""
+
+
+
 
 
 
@@ -588,12 +823,18 @@ class UnitsManaged(Managed):
     
     """    
     
-    def convert_energy_2_internal_u(self,val):
+    def convert_energy_2_internal_u(self, val):
         return self.manager.convert_energy_2_internal_u(val)
         
-    def convert_energy_2_current_u(self,val):
+    def convert_energy_2_current_u(self, val):
         return self.manager.convert_energy_2_current_u(val)
+ 
+    def convert_length_2_internal_u(self, val):
+        return self.manager.convert_length_2_internal_u(val)
         
+    def convert_length_2_current_u(self, val):
+        return self.manager.convert_length_2_current_u(val)    
+       
     def unit_repr(self,utype="energy"):
         return self.manager.unit_repr(utype)
         
@@ -615,9 +856,30 @@ class EnergyUnitsManaged(Managed):
     def unit_repr(self):
         return self.manager.unit_repr("energy")
 
-    def unit_repr_latex(self,utype="energy"):
+    def unit_repr_latex(self, utype="energy"):
         return self.manager.unit_repr_latex(utype)
-        
+    
+
+class LengthUnitsManaged(Managed):
+    """Class providing functions for length units conversion
+    
+    """
+    
+    utype = "length"
+    units = "A"
+    
+    def convert_2_internal_u(self, val):
+        return self.manager.convert_length_2_internal_u(val)
+ 
+    def convert_2_current_u(self,val):
+        return self.manager.convert_length_2_current_u(val)
+    
+    def unit_repr(self):
+        return self.manager.unit_repr(self.utype)
+
+    def unit_repr_latex(self):
+        return self.manager.unit_repr_latex(self.utype)
+    
         
 class BasisManaged(Managed):
     """Base class for objects with managed basis
@@ -641,6 +903,8 @@ class BasisManaged(Managed):
         
         
 
+
+
 class units_context_manager:
     """General context manager to manage physical units of values 
     
@@ -692,6 +956,30 @@ class frequency_units(energy_units):
     """
     pass
         
+
+class length_units(units_context_manager):
+    """Context manager for length units
+    
+    
+    """
+    
+    def __init__(self, units):
+        super().__init__(utype="length")
+        
+        if units in self.manager.units["length"]:
+            self.units = units
+        else:
+            raise Exception("Unknown length units")
+            
+    def __enter__(self):
+        # save current energy units
+        self.units_backup = self.manager.get_current_units("length")
+        self.manager.set_current_units(self.utype,self.units)
+        
+    def __exit__(self,ext_ty,exc_val,tb):
+        self.manager.set_current_units("length",self.units_backup)
+
+
         
 class basis_context_manager:
     """General context manager to manage basis 
@@ -719,6 +1007,7 @@ class eigenbasis_of(basis_context_manager):
     def __init__(self, operator):
         super().__init__()
         self.op = operator
+        self.manager.store_current_basis_operator(self.op)
         
         
     def __enter__(self):
@@ -736,7 +1025,7 @@ class eigenbasis_of(basis_context_manager):
         
         #SS = self.op.diagonalize()
         SS = self.op.get_diagonalization_matrix()
-        nb = self.manager.set_new_basis(SS)
+        self.manager.set_new_basis(SS)
 
         #self.manager.register_with_basis(nb,self.op)
         #self.op.set_current_basis(nb)
@@ -782,6 +1071,7 @@ class eigenbasis_of(basis_context_manager):
                 if op not in ops_above:
                     self.manager.register_with_basis(nb,op)
             
+        self.manager.remove_current_basis_operator()
             
         del self.manager.basis_registered[bb]
 
