@@ -1,25 +1,6 @@
 # -*- coding: utf-8 -*-
 """
-    Quantarhei package (http://www.github.com/quantarhei)
-
-    correlationfunctions module
-
-
-"""
-
-import numpy
-import scipy.interpolate as interp
-
-from ...core.dfunction import DFunction
-from ...core.units import kB_intK
-from ...core.managers import UnitsManaged
-from ...core.managers import energy_units
-from ...core.time import TimeAxis
-from ...core.frequency import FrequencyAxis
-
-
-class CorrelationFunction(DFunction, UnitsManaged):
-    """Provides typical Bath correlation function types.
+    Provides typical Bath correlation function types.
 
     Most important types of bath or energy gap correlation functions are
     provided. Where possible, the correlation function is calculated
@@ -27,55 +8,6 @@ class CorrelationFunction(DFunction, UnitsManaged):
     not available, correlation function is calculated by transformation
     of the spectral density.
 
-    Parameters
-    ----------
-
-    axis : TimeAxis
-        TimeAxis object specifying the time interval on which the
-        correlation function is defined.
-
-    params : dictionary
-        A dictionary of the correlation function parameters
-
-    values : optional
-        Correlation function can be set by specifying values at all times
-
-    Methods
-    -------
-
-    is_analytical()
-        Returns `True` if the correlation function is calculated from an
-        analytical formula, `False` otherwise.
-
-    copy()
-        Returns a copy of the CorrelationFunction object
-
-    get_temperature()
-        Returns the temperature of the correlation function
-
-    get_reorganization_energy()
-        Returns the reorganization energy parameters of
-        the correlation function
-
-    measure_reorganization_energy()
-        Calculates reorganization energy from the shape of the correlation
-        function
-
-    get_FTCorrelationFunction()
-        Returns the Fourier transform of the correlation function
-
-    get_EvenFTCorrelationFunction()
-        Returns the Fourier transform of the real part of the correlation
-        function
-
-    get_OddFTCorrelationFunction()
-        Returns the Fourier transform of the imaginary part of the correlation
-        function
-
-    get_SpectralDensity()
-        Returns numerically calculated spectral density
-
-
     Types of correlation function provided
     --------------------------------------
     OverdampedBrownian-HighTemperature :
@@ -107,17 +39,57 @@ class CorrelationFunction(DFunction, UnitsManaged):
     >>> print(numpy.allclose(lamb_definition, lamb_measured, rtol=1.0e-4))
     True
 
+    Details of Classes Provided
+    ---------------------------
+    
+"""
+
+import numpy
+import scipy.interpolate as interp
+
+from ...core.dfunction import DFunction
+from ...core.units import kB_intK
+from ...core.managers import UnitsManaged
+from ...core.managers import energy_units
+from ...core.time import TimeAxis
+from ...core.frequency import FrequencyAxis
+
+
+class CorrelationFunction(DFunction, UnitsManaged):
+    """Provides typical Bath correlation function types.
+
+
+    Parameters
+    ----------
+
+    axis : TimeAxis
+        TimeAxis object specifying the time interval on which the
+        correlation function is defined.
+
+    params : dictionary
+        A dictionary of the correlation function parameters
+
+    values : optional
+        Correlation function can be set by specifying values at all times
+
+
     """
 
     allowed_types = ("OverdampedBrownian-HighTemperature",
-                     "OverdampedBrownian", 
+                     "OverdampedBrownian",
+                     "OverdampedBrownian_from_Specdens",
                      "UnderdampedBrownian",
-                     "Value-defined")
+                     "Underdamped",
+                     "B777",
+                     "CP29",
+                     "Value-defined"
+                     )
 
     analytical_types = ("OverdampedBrownian-HighTemperature",
                         "OverdampedBrownian")
     
-    energy_params = ("reorg", "omega", "freq")
+    energy_params = ("reorg", "omega", "freq", "fcp", "g_FWHM", "l_FWHM",\
+                     "freq1", "freq2", "gamma")
 
     def __init__(self, axis=None, params=None , values=None):
         super().__init__()
@@ -153,10 +125,7 @@ class CorrelationFunction(DFunction, UnitsManaged):
             self.lamb = 0.0
             self.temperature = -1.0
             self.cutoff_time = 0.0
-    
-            #
-            # loop over parameter sets
-            #
+
             for params in p2calc:
                 
                 try:
@@ -176,28 +145,85 @@ class CorrelationFunction(DFunction, UnitsManaged):
                 except:
                     raise Exception("Dictionary of parameters does not contain "
                                     +" `ftype` key")
+                    
+                self.params.append(prms)                    
     
-                if ftype == "OverdampedBrownian-HighTemperature":
-        
-                    self._make_overdamped_brownian_ht(prms, values=values)
-        
-                elif ftype == "OverdampedBrownian":
-        
-                    self._make_overdamped_brownian(prms, values=values)
+    
+            if values is None:
+                #
+                # loop over parameter sets
+                #
+                for prms in self.params:
                     
-                elif ftype == "UnderdampedBrownian":
+#                    try:
+#                        ftype = params["ftype"]
+#                        
+#                        if ftype not in CorrelationFunction.allowed_types:
+#                            raise Exception("Unknown CorrelationFunction type")
+#            
+#                        # we mutate the parameters into internal units
+#                        prms = {}
+#                        for key in params.keys():
+#                            if key in self.energy_params:
+#                                prms[key] = self.convert_energy_2_internal_u(params[key])
+#                            else:
+#                                prms[key] = params[key]
+#                                
+#                    except:
+#                        raise Exception("Dictionary of parameters does not contain "
+#                                        +" `ftype` key")                    
+        
+                    if ftype == "OverdampedBrownian-HighTemperature":
+            
+                        self._make_overdamped_brownian_ht(prms) #, values=values)
+            
+                    elif ftype == "OverdampedBrownian":
+            
+                        self._make_overdamped_brownian(prms) #, values=values)
+                        
+                    elif ftype == "UnderdampedBrownian":
+                        
+                        self._make_underdamped_brownian(prms) #, values=values)
+                        
+                    elif ftype == "Underdamped":
+                        
+                        self._make_underdamped(params, values=values)
+                        
+                    elif ftype == "B777":
+                        
+                        self._make_B777(params, values=values)
+                        
+                    elif ftype == "CP29":
+                        
+                        self._make_CP29_spectral_density(params, values=values)
+            
+                    elif ftype == "Value-defined":
+            
+                        self._make_value_defined(prms, values)
+            
+                    else:
+                        raise Exception("Unknown correlation function type or"+
+                                        "type domain combination.")
+                        
+                    #self.params.append(prms)
                     
-                    self._make_underdamped_brownian(prms, values=values)
-        
-                elif ftype == "Value-defined":
-        
-                    self._make_value_defined(prms, values)
-        
-                else:
-                    raise Exception("Unknown correlation function type or"+
-                                    "type domain combination.")
+            else:
+                
+                self._add_me(self.axis, values) 
+                # update reorganization energy
+                self.lamb = 0.0
+                self.temperature = self.params[0]["T"]
+                for prms in self.params:
+                    self.lamb += prms["reorg"]
+                    if self.temperature != prms["T"]:
+                        raise Exception("Inconsistent temperature! "
+                                        +"Temperatures of all "
+                                        +"components have to be the same")
                     
-                self.params.append(prms)
+                #FIXME: set cut-off time and temperature
+                #self._set_temperature_and_cutoff_time(self.params[0])
+                    
+                
 
     def _matsubara(self, kBT, ctime, nof):
         """Matsubara frequency part of the Brownian correlation function
@@ -230,7 +256,7 @@ class CorrelationFunction(DFunction, UnitsManaged):
             self.cutoff_time = new_cutoff_time
             
             
-    def _make_overdamped_brownian(self, params, values=None):
+    def _make_overdamped_brownian(self, params): #, values=None):
         """Creates the overdamped Brownian oscillator component
         of the correlation function
         
@@ -248,10 +274,12 @@ class CorrelationFunction(DFunction, UnitsManaged):
         kBT = kB_intK*temperature
         time = self.axis.data
 
-        if values is not None:
-            cfce = values
+        #if values is not None:
+        #    cfce = values
+            
+        #else:
+        if True:
             
-        else:
             cfce = (lamb/(ctime*numpy.tan(1.0/(2.0*kBT*ctime))))\
                 *numpy.exp(-time/ctime) \
                 - 1.0j*(lamb/ctime)*numpy.exp(-time/ctime)
@@ -269,7 +297,7 @@ class CorrelationFunction(DFunction, UnitsManaged):
 
 
 
-    def _make_overdamped_brownian_ht(self, params, values=None):
+    def _make_overdamped_brownian_ht(self, params): # , values=None):
         """Creates the high temperature overdamped Brownian oscillator 
         component of the correlation function
         
@@ -281,10 +309,11 @@ class CorrelationFunction(DFunction, UnitsManaged):
         kBT = kB_intK*temperature
         time = self.axis.data
 
-        if values is not None:
-            cfce = values
-            
-        else:
+        #if values is not None:
+        #    cfce = values
+        #    
+        #else:
+        if True:
             cfce = 2.0*lamb*kBT*(numpy.exp(-time/ctime)
                                  - 1.0j*(lamb/ctime)*numpy.exp(-time/ctime))
 
@@ -298,7 +327,7 @@ class CorrelationFunction(DFunction, UnitsManaged):
         self._set_temperature_and_cutoff_time(temperature, 5.0*ctime)  
         
 
-    def _make_underdamped_brownian(self, params, values=None):
+    def _make_underdamped_brownian(self, params): #, values=None):
         """Creates underdamped Brownian oscillator component of the correlation
         function
         
@@ -314,10 +343,11 @@ class CorrelationFunction(DFunction, UnitsManaged):
         #kBT = kB_intK*temperature
         time = self.axis #.data
 
-        if values is not None:
-            cfce = values
-            
-        else:
+        #if values is not None:
+        #    cfce = values
+        #    
+        #else:
+        if True:
             with energy_units("int"):
                 # Make it via SpectralDensity
                 fa = SpectralDensity(time, params)
@@ -338,6 +368,99 @@ class CorrelationFunction(DFunction, UnitsManaged):
         self._set_temperature_and_cutoff_time(temperature, 5.0/ctime)  
         
         
+    def _make_underdamped(self, params, values=None):
+        from .spectraldensities import SpectralDensity
+        
+        temperature = params["T"]
+        ctime = params["gamma"]
+        
+        # use the units in which params was defined
+        lamb = params["reorg"]
+        time = self.axis #.data
+
+        if values is not None:
+            cfce = values
+        else:
+            # Make it via SpectralDensity
+            fa = SpectralDensity(time, params)
+            
+            cf = fa.get_CorrelationFunction(temperature=temperature)
+            
+            cfce = cf.data
+
+         # this is a call to the function inherited from DFunction class 
+        self._add_me(self.axis, cfce)
+
+        # update reorganization energy
+        self.lamb += lamb
+        
+        # check temperature and update cutoff time
+        self._set_temperature_and_cutoff_time(temperature, 5.0*ctime) 
+        
+    def _make_B777(self, params, values=None):
+        from .spectraldensities import SpectralDensity
+        
+        temperature = params["T"]
+        ctime = params["gamma"]
+        
+        # use the units in which params was defined
+        lamb = self.manager.iu_energy(params["reorg"],
+                                      units=self.energy_units)
+        time = self.axis #.data
+
+        if values is not None:
+            cfce = values
+        else:
+            # Make it via SpectralDensity
+            fa = SpectralDensity(time, params)
+            
+            cf = fa.get_CorrelationFunction(temperature=temperature)
+            
+            cfce = cf.data
+            
+        # this is a call to the function inherited from DFunction class 
+        self._add_me(self.axis, cfce)
+
+        # update reorganization energy
+        self.lamb += lamb
+        
+        # check temperature and update cutoff time
+        self._set_temperature_and_cutoff_time(temperature, 5.0*ctime)     
+
+        
+    def _make_CP29_spectral_density(self, params, values=None):
+        from .spectraldensities import SpectralDensity
+        
+        temperature = params["T"]
+        ctime = params["gamma"]
+        #omega = params["freq"]
+        
+        # use the units in which params was defined
+        lamb = self.manager.iu_energy(params["reorg"],
+                                      units=self.energy_units)
+        print('correlation function lamb in int units %f' %lamb)
+        time = self.axis #.data
+
+        if values is not None:
+            cfce = values
+        else:
+            # Make it via SpectralDensity
+            fa = SpectralDensity(time, params)
+            
+
+            cf = fa.get_CorrelationFunction(temperature=temperature)
+            
+            cfce = cf.data
+
+        # this is a call to the function inherited from DFunction class 
+        self._add_me(self.axis, cfce)
+
+        # update reorganization energy
+        self.lamb += lamb
+        
+        # check temperature and update cutoff time
+        self._set_temperature_and_cutoff_time(temperature, 5.0*ctime)   
+        
     def _make_value_defined(self, params, values):
         
         lamb = params["reorg"]
@@ -469,6 +592,7 @@ class CorrelationFunction(DFunction, UnitsManaged):
         else:
             return False
 
+
     def is_analytical(self):
         """Returns `True` if analytical
 
@@ -486,12 +610,19 @@ class CorrelationFunction(DFunction, UnitsManaged):
         """
         return self.temperature
 
+
     def get_reorganization_energy(self):
         """Returns the reorganization energy of the correlation function
 
         """
         return self.convert_energy_2_current_u(self.lamb)
 
+    def get_correlation_time(self):
+        """Returns correlation time associated with the first component 
+        of the bath correlation function
+        """
+        return self.params[0]["ctime"]
+
     def measure_reorganization_energy(self):
         """Calculates the reorganization energy of the correlation function
 
@@ -514,11 +645,12 @@ class CorrelationFunction(DFunction, UnitsManaged):
         return cfce
 
 
-    def get_SpectralDensity(self):
-        """ Returns a corresponding SpectralDensity object
-
-        Returns a SpectralDensity corresponding to this CorrelationFunction
-
+    def get_SpectralDensity(self, fa=None):
+        """
+        Returns a SpectralDensity corresponding to this CorrelationFunction.
+        If a FrequencyAxis object is included, the SpectralDensity
+        object will be returned with that FrequencyAxis instance as its 
+        frequency axis.
 
         """
 
@@ -528,6 +660,14 @@ class CorrelationFunction(DFunction, UnitsManaged):
         with energy_units("int"):
             frequencies = self.axis.get_FrequencyAxis()
             vals = self.get_OddFTCorrelationFunction().data
+            
+            if fa is not None:
+                if numpy.all(numpy.isclose(fa.data, frequencies.data, 1e-5)):
+                    time = ta
+                else:
+                    raise Exception("The provided FrequencyAxis does not "
+                                    + "have the same data as the Fourier "
+                                    + "transformed axis")
 
             # FIXME: how to set the limit of SpectralDensity at w->0
             spectd = SpectralDensity(frequencies, self.params, values=vals)
@@ -546,6 +686,7 @@ class CorrelationFunction(DFunction, UnitsManaged):
             ftcf = FTCorrelationFunction(self.axis, self.params)
         return ftcf
 
+
     def get_OddFTCorrelationFunction(self):
         """Returns a odd part of the Fourier transform of correlation function
 
@@ -559,6 +700,7 @@ class CorrelationFunction(DFunction, UnitsManaged):
             oftcf = OddFTCorrelationFunction(self.axis, self.params)
         return oftcf
 
+
     def get_EvenFTCorrelationFunction(self):
         """Returns a even part of the Fourier transform of correlation function
 
